//
//  SyncDatabase.m
//  TicTacToe
//
//  Created by Kasprzak Ingo on 28.06.12.
//  Copyright (c) 2012 Silutions. All rights reserved.
//

#import "SyncDatabase.h"
@import ObjectiveC;

#import "Device+Methods.h"
#import "SynchronizationInfo+Methods.h"
#import "DeletionInfo+Methods.h"

#import "SynchronizationServer.h"
#import "MultipeerConnectivitySynchronizationServer.h"

// Keys for synchronization dictionary
#define SyncDatabaseSyncCommandKey @"command"
#define SyncDatabaseSyncDataKey @"data"
#define SyncDatabaseSyncDateKey @"date"
#define SyncDatabaseSyncSenderUIDKey @"senderUID"
#define SyncDatabaseSyncDataUIDKey @"dataUID"
#define SyncDatabaseSyncWantsReceiptKey @"wantsReceipt"


@interface SyncDatabase()
//@property (strong) NSMutableDictionary *unconfirmedCommands;
//@property (strong) NSString *remotePeerID;
//@property (strong, readwrite) NSString *remoteDeviceUID;
@property (strong) SynchronizationServer *synchronizationServer;
@end

@implementation SyncDatabase

#pragma mark - Initialization


+ (void) initialize {
	if ([self class] == [SyncDatabase class] && [NSManagedObject shouldReturnNo]) {
		Method shouldReturnYes_original = class_getClassMethod([NSManagedObject class], @selector(shouldReturnYes));
		Method shouldReturnNo_original =  class_getClassMethod([NSManagedObject class], @selector(shouldReturnNo));
		method_exchangeImplementations(shouldReturnYes_original, shouldReturnNo_original);
		
		// Some method swizzling on NSManagedObject
		Method awakeFromInsert_custom = class_getInstanceMethod([NSManagedObject class], @selector(awakeFromInsert_custom));
		Method awakeFromInsert_original = class_getInstanceMethod([NSManagedObject class], @selector(awakeFromInsert));
		method_exchangeImplementations(awakeFromInsert_original, awakeFromInsert_custom);
		
		Method willSave_custom = class_getInstanceMethod([NSManagedObject class], @selector(willSave_custom));
		Method willSave_original = class_getInstanceMethod([NSManagedObject class], @selector(willSave));
		method_exchangeImplementations(willSave_original, willSave_custom);
		
		Method prepareForDeletion_custom = class_getInstanceMethod([NSManagedObject class], @selector(prepareForDeletion_custom));
		Method prepareForDeletion_original = class_getInstanceMethod([NSManagedObject class], @selector(prepareForDeletion));
		method_exchangeImplementations(prepareForDeletion_original, prepareForDeletion_custom);
		
		Method save_custom = class_getInstanceMethod([NSManagedObjectContext class], @selector(save_custom:));
		Method save_original = class_getInstanceMethod([NSManagedObjectContext class], @selector(save:));
		method_exchangeImplementations(save_original, save_custom);
		NSLog(@"Method swizzling initiated");
	}
}

- (void)startSynchronizationServer {
    if (!self.synchronizationServer) {
		self.synchronizationServer = [[MultipeerConnectivitySynchronizationServer alloc] init];
		self.synchronizationServer.delegate = self;
		[self.synchronizationServer startServer];

		[[NSNotificationCenter defaultCenter] addObserver:self
												 selector:@selector(managedObjectContextDidSave:)
													 name:NSManagedObjectContextDidSaveNotification
												   object:self.managedObjectContext];
	}
}

- (void)stopSynchronizationServer {
	if (self.synchronizationServer) {
		[self.synchronizationServer stopServer];
		self.synchronizationServer = nil;

		[[NSNotificationCenter defaultCenter] removeObserver:self
														name:NSManagedObjectContextDidSaveNotification
													  object:self.managedObjectContext];
	}
}

- (Device*)thisDevice { // returns data set for this device
	Device *result = nil;
	
	NSString *myUID = [NSManagedObject thisDeviceUid];

	if (myUID) {
        result = [self fetchObjectForEntityClass:[Device class]
                                   withPredicate:@"%K=%@", ManagedObjectUIDKey, myUID];
	}

	if (!result) {
        result = [NSEntityDescription insertNewObjectForEntityForName:NSStringFromClass([Device class])
                                               inManagedObjectContext:self.managedObjectContext];
        // setting of properties for entity in category NSManagedObject+UID: - (void)awakeFromInsert_custom
	}
    
	return result;
}

- (NSArray*)connectedDeviceUIDs {
	return [self.synchronizationServer connectedDeviceUIDs];
}

#pragma mark - Synchronization methods

- (NSDate*)synchronizationDateForDevice:(NSString*)deviceUID {
	NSDate *result = nil;
	NSSet *set = [self fetchObjectsForEntityClass:[SynchronizationInfo class]
									withPredicate:@"%K=%@", SynchronizationInfoDeviceUIDKey, deviceUID];
	if ([set count]) {
		result = [[set anyObject] lastSynchronization];
	} else {
		SynchronizationInfo *syncInfo = [NSEntityDescription insertNewObjectForEntityForName: NSStringFromClass([SynchronizationInfo class])
																	  inManagedObjectContext:self.managedObjectContext];
		syncInfo.deviceUID = deviceUID;
		result = [NSDate dateWithTimeIntervalSince1970:0];
		syncInfo.lastSynchronization = result;
	}
	return result;
}

- (void)updateSynchronizationDateForDevice:(NSString*)deviceUID {
	NSSet *set = [self fetchObjectsForEntityClass:[SynchronizationInfo class]
									withPredicate:@"%K=%@", SynchronizationInfoDeviceUIDKey, deviceUID];
	if ([set count]) {
		[[set anyObject] setLastSynchronization:[NSDate date]];
	} else {
		SynchronizationInfo *syncInfo = [NSEntityDescription insertNewObjectForEntityForName: NSStringFromClass([SynchronizationInfo class])
																	  inManagedObjectContext:self.managedObjectContext];
		syncInfo.deviceUID = deviceUID;
		syncInfo.lastSynchronization = [NSDate date];
	}
}

- (void)synchronizeDeviceBidirectional:(BOOL)bidirectional {
	NSArray *entities = [self.managedObjectModel entities];
	
	NSArray *deviceUIDs = self.connectedDeviceUIDs;
	for (NSString *deviceUID in deviceUIDs) {
		// Änderungen einsammeln
		NSDate *lastSyncDate = [self synchronizationDateForDevice:deviceUID];
		NSMutableSet *modifiedObjectsSet = [NSMutableSet setWithCapacity:50];
        // Suchen aller nach dem letzten Sync geänderten managed objects über alle entities
		for (NSEntityDescription *entity in entities) {
            // SynchronizationInfo ist "private" Entity
			if (![entity.managedObjectClassName isEqualToString:NSStringFromClass([SynchronizationInfo class])]) {
				NSSet *set = [self fetchObjectsForEntityClass:NSClassFromString(entity.managedObjectClassName)
												withPredicate:@"%K>=%@ AND %K!=%@",
							  ManagedObjectModificationDateKey, lastSyncDate,
							  ManagedObjectModificationDeviceUid, deviceUID];
				[set enumerateObjectsUsingBlock:^(id obj, BOOL *stop) {
					[modifiedObjectsSet addObject:[obj propertyDictionary]];
				}];
			}
		}
        
        // Die gesammelten Änderungen an das Gerät versenden.
		if ([modifiedObjectsSet count]) {
			NSString *command = bidirectional ? SyncDatabaseSyncSynchronizeBidirectionalCommand : SyncDatabaseSyncSynchronizeCommand;
			
			[self.synchronizationServer sendObject:modifiedObjectsSet
									   withCommand:command
										 toDevices:@[deviceUID]
									  wantsReceipt:YES];
		}
	}
}

// Update or insert a single managed object with the values in the propertyDictionary
// YOU HAVE TO SET [NSManagedObject setSynchronizing:YES] BEFORE USING THIS METHOD!
- (NSManagedObject*)updateDatabaseWithPropertyDictionary:(NSDictionary*)propertyDictionary {
	
	NSString *entityName = [propertyDictionary objectForKey:PropertyDictionaryEntityNameKey];
	NSString *uid = [propertyDictionary objectForKey:ManagedObjectUIDKey];
	NSSet *set = [self fetchObjectsForEntityClass:NSClassFromString(entityName)
									withPredicate:@"%K = %@", ManagedObjectUIDKey, uid];
	NSManagedObject *object = nil;
	if (![set count]) {
		object = [NSEntityDescription insertNewObjectForEntityForName:entityName
											   inManagedObjectContext:self.managedObjectContext];
	} else {
		object = [set anyObject];
	}
	
    // ---
	NSDictionary *properties = object.entity.propertiesByName;
	[propertyDictionary enumerateKeysAndObjectsUsingBlock:^(NSString* key, id value, BOOL *stop)
    {
		id property = [properties objectForKey:key];
        
		if (property) {
			if (value == [NSNull null]) { value = nil; }
			
            if (value && [property isKindOfClass:[NSRelationshipDescription class]])
            {
				NSFetchRequest *request = [NSFetchRequest fetchRequestWithEntityName:[[property destinationEntity] name]];
				request.predicate = [NSPredicate predicateWithFormat:@"%K=%@", ManagedObjectUIDKey, value];
				NSArray *result = [self.managedObjectContext executeFetchRequest:request error:nil];
				
                if ([result count]) {       // NSManagedObject existiert
					value = [result lastObject];
				} else {                    // NSManagedObject existiert noch nicht
					NSManagedObject *managedObj =
                        [NSEntityDescription insertNewObjectForEntityForName:[[property destinationEntity] name]
                                                      inManagedObjectContext:self.managedObjectContext];
					[managedObj setValue:value forKey:ManagedObjectUIDKey];
					value = managedObj;
				}
			}
			[object setValue:value forKey:key];
		}
	}];

	return object;
}

- (void)managedObjectContextDidSave:(NSNotification*)notification {
	if (self.synchronizationServer && [self.connectedDeviceUIDs count]) {
		[self synchronizeDeviceBidirectional:NO];
	}
}

#pragma mark - Synchronization Server

- (void)receivedModifiedObjects:(NSSet*)objects {
	DBLog(@"Received %i objects.", [objects count]);
    
    // Flag setzen, dass unsere eigene Intelligenz zurzeit nicht gefragt ist
	[NSManagedObject setSynchronizing:YES];

	// Update/insert objects
	[objects enumerateObjectsUsingBlock:^(id obj, BOOL *stop) {
		[self updateDatabaseWithPropertyDictionary:obj];
	}];
	
	// Delete objects
	NSSet *deletions = [objects filteredSetUsingPredicate: [NSPredicate predicateWithFormat:@"%K==%@",
															PropertyDictionaryEntityNameKey,
															NSStringFromClass([DeletionInfo class])]];
	__block int count = 0;
	[deletions enumerateObjectsUsingBlock:^(NSDictionary *obj, BOOL *stop) {
		NSManagedObject *managedObj =
		[self fetchObjectForEntityClass:NSClassFromString([obj objectForKey:DeletionInfoEntityTypeKey])
						  withPredicate:@"%K==%@", ManagedObjectUIDKey, [obj objectForKey:DeletionInfoObjectUidKey]];
		if (managedObj) {
			[self.managedObjectContext deleteObject:managedObj];
			count++;
		}
	}];
	if (count) { DBLog(@"Deleted %i objects.", count); }
	
	[self saveChanges];
	[NSManagedObject setSynchronizing:NO];
}

- (void)receivedBidirectionalRequestAndModifiedObjects:(NSSet*)objects {
	
	[self receivedModifiedObjects:objects];
	
	DBLog(@"Sending back synchronization data...");
	[self synchronizeDeviceBidirectional:NO];
}

- (void)receivedReceiptForCommand:(NSString *)commandString
					   fromDevice:(NSString *)deviceUID
						 withDate:(NSDate *)date {
	if ([commandString isEqualToString:SyncDatabaseSyncSynchronizeCommand]) {
		[self updateSynchronizationDateForDevice:deviceUID];
		DBLog(@"Synchronisation date for device '%@' updated.", deviceUID);
	}
}

- (void)receivedIdentification:(NSString*)deviceUID fromPeer:(NSString*)peer {
	[self synchronizeDeviceBidirectional:NO];
	[[NSNotificationCenter defaultCenter] postNotificationName:SyncDatabaseConnectedToNewDeviceNotification
														object:self];
}

- (void)receivedDisconnection:(NSString*)deviceUID fromPeer:(NSString*)peer {
	[[NSNotificationCenter defaultCenter] postNotificationName:SyncDatabaseDisconnectedDeviceNotification
														object:self];
}


@end
