//
//  ViewController.m
//  TicTacToe
//
//  Created by Kasprzak Ingo on 28.06.12.
//  Copyright (c) 2012 Silutions. All rights reserved.
//

#import "ViewController.h"
#import "Move+Methods.h"
#import "Game+Methods.h"
#import "Device+Methods.h"
#import "TableViewController.h"

@interface ViewController ()

@end

@implementation ViewController
@synthesize startNewGame;
@synthesize fields;
@synthesize localName;
@synthesize remoteName;


#pragma mark - View life cycle

- (void)viewDidLoad {
    [super viewDidLoad];

	// add actions to game field buttons
	[self.fields enumerateObjectsUsingBlock:^(UIButton *button, NSUInteger idx, BOOL *stop) {
		[button addTarget:self
				   action:@selector(touchField:)
		 forControlEvents:UIControlEventTouchUpInside];
	}];

	// Notifications
	[[NSNotificationCenter defaultCenter] addObserver:self
											 selector:@selector(managedObjectContextDidSave:)
												 name:NSManagedObjectContextDidSaveNotification
											   object:[[TicTacToeDatabase sharedTicTacToeDatabase] managedObjectContext]];

	[[NSNotificationCenter defaultCenter] addObserver:self
											 selector:@selector(connectedToNewDeviceNotification:)
												 name:SyncDatabaseConnectedToNewDeviceNotification
											   object:nil];

	[[NSNotificationCenter defaultCenter] addObserver:self
											 selector:@selector(disconnectedDeviceNotification:)
												 name:SyncDatabaseDisconnectedDeviceNotification
											   object:nil];

	// KVO
	[[TicTacToeDatabase sharedTicTacToeDatabase] addObserver:self
												  forKeyPath:@"remoteDeviceUID"
													 options:NSKeyValueObservingOptionNew context:nil];

	// update play field views
	[self updatePlayfield];
	
	self.localName.text = [[[TicTacToeDatabase sharedTicTacToeDatabase] thisDevice] playerName];
}

- (void)viewDidUnload {
	[[NSNotificationCenter defaultCenter] removeObserver:self];
	[[TicTacToeDatabase sharedTicTacToeDatabase] removeObserver:self forKeyPath:@"remoteDeviceUID"];
	
    [self setStartNewGame:nil];
	[self setFields:nil];
	
	[self setLocalName:nil];
	[self setRemoteName:nil];
    
    [super viewDidUnload];
}


#pragma mark - Private

- (void)updatePlayfield { // Spielfeld-Views updaten
	TicTacToeDatabase *database = [TicTacToeDatabase sharedTicTacToeDatabase];
	Game *game = database.activeGame;
	
    // Step 1: erase all play fields
	NSSet *winMoves = [game checkForWinningCombo];
	[self.fields enumerateObjectsUsingBlock:^(UIButton *button, NSUInteger idx, BOOL *stop) {
		button.enabled = (winMoves==nil);
		[button setTitle:@"" forState:UIControlStateNormal];
	}];
	
    // Step 2: update play fields with game history
	NSSet *moves = game.moves;
	[moves enumerateObjectsUsingBlock:^(Move *obj, BOOL *stop) {
		UIButton *field = [[self.fields filteredArrayUsingPredicate:
							[NSPredicate predicateWithFormat:@"tag=%@", obj.field]]
						   lastObject];
		field.enabled = NO;
		UIColor *color = (winMoves && [winMoves containsObject:obj]) ? [UIColor redColor] : [UIColor darkTextColor];
		[field setTitleColor:color forState:UIControlStateNormal];
		[field setTitle: ([obj.player intValue] ? @"O" : @"X")
			   forState: UIControlStateNormal];
	}];
    
    // Step 3: Update remote player name
	NSArray *remoteUIDs = database.connectedDeviceUIDs;
	if ([remoteUIDs count]) {
		Device *remoteDevice = [database fetchObjectForEntityClass:[Device class]
													 withPredicate:@"%K=%@", ManagedObjectUIDKey, [remoteUIDs lastObject]];
		if (remoteDevice) {
			NSString *temp = remoteDevice.playerName;
			if ([temp length]==0) { temp = @"No name"; }
			self.remoteName.text = temp;
		}
	} else {
		self.remoteName.text = @"No remote";
	}
}

#pragma mark - Actions

- (IBAction)touchStartNewGame:(id)sender {
	TicTacToeDatabase *database = [TicTacToeDatabase sharedTicTacToeDatabase];
	[database newGame];
	[database saveChanges];
}

- (IBAction)touchHistory:(id)sender {	
	NSArray *topLevelObjects = [[NSBundle mainBundle] loadNibNamed:@"NavigationViewController" owner:nil options:nil];
	[self presentViewController:[topLevelObjects lastObject] animated:YES completion:^{}];
}

- (void)touchField:(UIButton*)button {
	TicTacToeDatabase *database = [TicTacToeDatabase sharedTicTacToeDatabase];

	Move *move = [database newMoveForActiveGame];
	move.field = @([button tag]);
    
	[database saveChanges];
}

- (BOOL)textFieldShouldReturn:(UITextField *)textField {
	TicTacToeDatabase *database = [TicTacToeDatabase sharedTicTacToeDatabase];

	[textField resignFirstResponder];
	
    Device *device = database.thisDevice;
	device.playerName = [self.localName text];
	
    [database saveChanges];
	
    return YES;
}

#pragma mark - Notifications

- (void)managedObjectContextDidSave:(NSNotification*)notification {
	[self updatePlayfield];
}

- (void)connectedToNewDeviceNotification:(NSNotification*)notification {
	[self updatePlayfield];
}

- (void)disconnectedDeviceNotification:(NSNotification*)notification {
	[self updatePlayfield];
}

#pragma mark - KVO

- (void)observeValueForKeyPath:(NSString *)keyPath
					  ofObject:(id)object
						change:(NSDictionary *)change
					   context:(void *)context {
	if (object == [TicTacToeDatabase sharedTicTacToeDatabase]) {
		[self updatePlayfield];
	} else {
		[super observeValueForKeyPath:keyPath ofObject:object change:change context:context];
	}
}

#pragma mark - View notifications

- (NSUInteger)supportedInterfaceOrientations {
	return UIInterfaceOrientationMaskPortrait|UIInterfaceOrientationMaskPortraitUpsideDown;
}

- (UIInterfaceOrientation)preferredInterfaceOrientationForPresentation {
	return UIInterfaceOrientationPortrait;
}

@end
